package edu.georgetown.gucs.dictionary;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.*;
import java.util.Map.Entry;
import java.util.concurrent.LinkedBlockingQueue;
import edu.georgetown.gucs.utility.*;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.IllegalDataException;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import org.slf4j.LoggerFactory;


/**
 * Creates a list of unique tokens extracted from a collection of documents that can be trimmed by removing tokens based
 * on various different attributes; used for creating fingerprints of documents that are based on words that appear in a
 * document collection
 *
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class Dictionary implements Serializable, Iterable<Entry<Token, DictionaryEntry>> {

    private static final long serialVersionUID = 8334786910169166049L;
    private boolean trimmed;
    private int totalDocuments;
    private int maxThread;
    private int nextPosition;
    private double maxIDFfound;
    private double trimmed_max = 1.0;
    private double trimmed_min = 0.0;
    private String creatingProgram;
    private String creation;
    private String dictionaryFileName;
    private String dictionaryName;
    private String directory;
    private String GUID;
    private String systemID;
    private Map<Token, DictionaryEntry> dictionary = new HashMap<>();
    private TokenizerList tokenizers;


    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(Dictionary.class);


    /**
     * Constructor that sets default values for this dictionary
     */
    public Dictionary() throws SdTextException {
        try {
            this.dictionaryName = "";
            this.nextPosition = 0;
            this.totalDocuments = 0;
            this.maxIDFfound = 0.0;
            setCreatingProgram();
            setSystemID();
            this.GUID = new BigInteger(128, new Random()).toString();
            this.maxThread = Runtime.getRuntime().availableProcessors();
            this.trimmed = false;
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Constructor that loads this dictionary from a file
     *
     * @param dictionaryFilename the filename containing a dictionary
     */
    public Dictionary(String dictionaryFilename) throws SdTextException {
        try {
            this.loadDictionaryXML(dictionaryFilename);
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Loads a previously serialized XML dictionary object into this dictionary
     *
     * @param filename the string filename of the XML dictionary file to load
     */
    @SuppressWarnings("unchecked")
    private void loadDictionaryXML(String filename) throws SdTextException {
        this.dictionaryFileName = filename;

        try {
            File xml = new File(this.dictionaryFileName);
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(xml);
            Element dict = doc.getRootElement();

            this.dictionaryName = dict.getAttributeValue("name");


            this.GUID = dict.getAttributeValue("GUID");


            this.systemID = dict.getAttributeValue("systemID");


            this.creatingProgram = dict.getAttributeValue("creatingProgram");


            this.creation = dict.getAttributeValue("creation");
            this.directory = dict.getAttributeValue("directory");


            this.maxIDFfound = Double.valueOf(dict.getAttributeValue("maxIDF"));


            this.totalDocuments = Integer.parseInt(dict.getAttributeValue("totalDocuments"));


            this.trimmed = Boolean.valueOf(dict.getAttributeValue("trimmed"));


            if (this.trimmed) {

                this.trimmed_min = Double.valueOf(dict.getAttributeValue("trimmed_min"));
                this.trimmed_max = Double.valueOf(dict.getAttributeValue("trimmed_max"));

            }

            //findbugs todo test this
            synchronized (this) {
                this.nextPosition = Integer.parseInt(dict.getAttributeValue("nextPosition"));
            }


            Vector<String> tokenizerNames = new Vector<>();
            for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
                String tokenizer = tokenizerElement.getAttributeValue("name");
                if (tokenizerElement.getAttribute("parameter") != null) {
                    tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
                }
                tokenizerNames.add(tokenizer.trim());
            }
            setTokenizers(tokenizerNames);


            if (!this.dictionary.isEmpty()) {
                this.dictionary.clear();
            }

            for (Element tokenElement : (List<Element>) dict.getChildren("token")) {
                Token token = new Token(tokenElement.getAttributeValue("name"), Integer.parseInt(tokenElement
                        .getAttributeValue("start_byte")), Integer.parseInt(tokenElement.getAttributeValue("end_byte")));
                this.dictionary.put(
                        token,
                        new DictionaryEntry(token, Integer.parseInt(tokenElement.getAttributeValue("position")), Integer
                                .parseInt(tokenElement.getAttributeValue("frequency")), Integer.parseInt(tokenElement
                                .getAttributeValue("documentCount")), Double.parseDouble(tokenElement.getAttributeValue("idf")))
                );
            }

        } catch (FileNotFoundException e) {
            logger.error("File Not Found");
            ExceptionHandler.handleException(new Exception("File could not be found"));
        } catch (UnsupportedEncodingException e) {
            ExceptionHandler.handleException(new Exception("Unsupported Encoding"));
            logger.error("Unsuported Encoding");
        } catch (JDOMException e) {
            ExceptionHandler.handleException(new Exception("JDOM Exception"));
            logger.error("JDOM Exception");
        } catch (IOException e) {
            ExceptionHandler.handleException(new Exception("IO Exception"));
            logger.error("IOException");
        }
    }

    /**
     * Reads an XML file that specifies which tokenizers to use for this dictionary
     *
     * @param config_file the string name of the XML configuration file containing this dictionary's tokenizers
     */
    @SuppressWarnings("unchecked")
    private void addTokenizers(String config_file) throws SdTextException {


        try {
            File xml = new File(config_file);
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(xml);

            Element dict = doc.getRootElement();

            Vector<String> tokenizerNames = new Vector<>();

            for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
                String tokenizer = tokenizerElement.getAttributeValue("name");


                if (tokenizerNames.isEmpty() && !tokenizer.contains("FileTokenizer")) {

                    logger.error("FileTokenizer is not the first tokenizer in the list ");
                    ExceptionHandler.handleException(new Exception("Dictionary: First tokenizer (" + tokenizer
                            + ") must read from a file."));
                }

                if (tokenizerElement.getAttribute("parameter") != null) {
                    tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
                }
                tokenizerNames.add(tokenizer.trim());

            }
            setTokenizers(tokenizerNames);


        } catch (JDOMException | IOException e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Creates a deep copy of the given dictionary object and stores it in this dictionary
     *
     * @param anotherDictionary the dictionary object to copy
     */
    public void clone(Dictionary anotherDictionary) {
        this.dictionary.clear();
        this.dictionary.putAll(anotherDictionary.dictionary);
        this.directory = anotherDictionary.directory;

        this.nextPosition = anotherDictionary.nextPosition;

        this.trimmed = anotherDictionary.trimmed;

        if (this.trimmed) {
            this.trimmed_min = anotherDictionary.trimmed_min;


            this.trimmed_max = anotherDictionary.trimmed_max;

        }
        this.maxIDFfound = anotherDictionary.maxIDFfound;

        this.totalDocuments = anotherDictionary.totalDocuments;

        this.tokenizers = anotherDictionary.tokenizers;

    }

    /**
     * Determines if this dictionary contains a particular token
     *
     * @param token the string containing the token value
     * @return the <code>boolean</code> value of whether the token is present in this dictionary
     */
    public boolean containsToken(Token token) {
        return this.dictionary.containsKey(token);
    }

    /**
     * Creates this dictionary using the entire fileLister provided
     *
     * @param fileLister the fileLister containing the files to use for this dictionary
     */
    private void createFullDictionary(FileLister fileLister) throws SdTextException {
        try {

            this.dictionary.clear();

            this.directory = fileLister.getDirectory().getAbsolutePath();

	    if (fileLister.size() < Global.minimumFilesForDictionary){
		throw new SdTextException("Dictionary:CreateFullDictionary:Too few files to create dictionary, minimum set to " + Global.minimumFilesForDictionary + " in Global.minimumFilesForDictionary");
	    }

		
	    
            ArrayList<DictionaryWorker> threads = new ArrayList<>(maxThread);
            File poison = new File(new BigInteger(128, new Random()).toString());

            for (int i = 0; i < maxThread; i++) {
                threads.add(i, new DictionaryWorker(this.tokenizers.getNames(), this, fileLister, poison));
                threads.get(i).setName("Dictionary Worker " + i);
                threads.get(i).setDaemon(true);
                threads.get(i).start();
            }

            // Wait for the file lister to finish loading files if needed, but it shouldn't be

            while (!fileLister.getLoaded()) {
                try {
                    Thread.sleep(250);
                } catch (InterruptedException e) {
                    ExceptionHandler.handleException(new Exception("Create Full Dictionary: Interupted exception" +
                            "loading fileLister"));
                }
            }


            // Now wait for the dictionary worker threads to finish
            boolean done = false;
	    Integer debug_wait_count = 0;
            while (!done) {
                // Sleep a bit while the threads run
                try {
                    Thread.sleep(100);
		    debug_wait_count++;
		    if (debug_wait_count == 100){
			debug_wait_count = 0;
		    }
		    
                } catch (InterruptedException e) {
                    ExceptionHandler.handleException(new Exception("Create Full Dictionary: Interrupted Exception, " +
                            "threads improperly terminated "));
                }


                int waitingThreads = 0;
                for (int i = 0; i < threads.size(); i++) {
                    if (threads.get(i).getState() == Thread.State.valueOf("WAITING")) {
                        waitingThreads++;
                    }
                }

                // If so, then we are done loading.
                if (waitingThreads == threads.size()) {
                    // Pass the poison to kill the worker threads
                    fileLister.putPoison(poison);
                    done = true;
                }

            }

            // System.out.println("Workers completed");


            computeIDFs();
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Creates this dictionary using the optional count or percent in the given fileLister; if no count or percent is
     * specified then uses entire fileLister to create dictionary
     * <p>
     * fileLister the fileLister containing the files to use for this dictionary, which needs to have completed loading
     *
     * @param dictionaryFiles a list of files that will be used to create the dictionary
     * @param directory       the directory used to create the dictionary
     */
    public void createPartialDictionary(List<File> dictionaryFiles, File directory) throws SdTextException {
        try {
            if ((dictionaryFiles == null) || (dictionaryFiles.size() == 0)) {


                throw new RuntimeException("Dictionary: createPartialDictionarary - nullFileLister received, exiting");
            }


            this.dictionary.clear();

            this.directory = directory.getAbsolutePath();

            Iterator<File> dictionary_iterator = dictionaryFiles.iterator();

            LinkedBlockingQueue<File> sharedQueue = new LinkedBlockingQueue<>(128);

            File poison = new File("");

            // Create PartialDictionaryWorkerThreads to process the files
            ArrayList<PartialDictionaryWorker> threads = new ArrayList<>(maxThread);


            for (int i = 0; i < maxThread; i++) {
                threads.add(i, new PartialDictionaryWorker(this.tokenizers.getNames(), this, sharedQueue, poison));
                threads.get(i).setName("PartialDictionaryWorker " + i);
                threads.get(i).setDaemon(true);
                threads.get(i).start();
            }

            // Put files on the queue for the workers to process
            while (dictionary_iterator.hasNext()) {
                try {
                    File nextFile = dictionary_iterator.next();


                    sharedQueue.put(nextFile);
                } catch (InterruptedException e) {

                    ExceptionHandler.handleException(e);
                }
            }

            // put the poison in the queue
            try {
                sharedQueue.put(poison);
            } catch (InterruptedException e) {
                ExceptionHandler.handleException(e);
            }


            // wait a while for all threads to be waiting for more files, then interrupt them so they stop
            boolean done = false;
            while (!done) {

                try {


                    Thread.sleep(100);


                } catch (InterruptedException e) {
                    ExceptionHandler.handleException(e);
                }


                // Check if the queue is empty and all threads in waiting state
                int waitingThreads = 0;
                for (int i = 0; i < threads.size(); i++) {
                    if (threads.get(i).getState() == Thread.State.valueOf("TERMINATED")) {
                        waitingThreads++;
                    }
                }

                if (waitingThreads == threads.size()) {
                    done = true;
                }

            }

            computeIDFs();
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Allows for operations on the tokens in this dictionary
     *
     * @return the iterator over the tokens and their dictionaryEntry in this dictionary
     */
    @Override
    public synchronized Iterator<Entry<Token, DictionaryEntry>> iterator() {
        return this.dictionary.entrySet().iterator();
    }

    /**
     * Computes the Inverse Document Frequencies (IDFs) for each token in the dictionary and determines the largest IDF
     */
    private void computeIDFs() throws SdTextException {
        int count = 0;
        Set<Entry<Token, DictionaryEntry>> theList = this.dictionary.entrySet();
        try {
            for (Entry<Token, DictionaryEntry> token : theList) {
                if (token == null) {

                    logger.error("Null TOKEN!");
                    throw new RuntimeException("Null Token");
                }
                double idf = token.getValue().computeIDF(this.totalDocuments);
                this.maxIDFfound = (idf > this.maxIDFfound) ? idf : this.maxIDFfound;
            }
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }


    }

    /**
     * Provides the program that created this dictionary
     *
     * @return the string name of the program that created this dictionary
     */
    public String getCreatingProgram() {
        return this.creatingProgram;
    }

    /**
     * Provides the maximum thread count to use for this dictionary
     *
     * @return the maximum thread count to use for this dictionary
     */
    public int getMaxThread() {
        return this.maxThread;
    }

    /**
     * Provides the original creation date of this dictionary
     *
     * @return the string date when this dictionary was created
     */
    public String getCreation() {
        return this.creation;
    }

    /**
     * Provides the filename of this dictionary, if this dictionary was loaded from or saved to a file
     *
     * @return the string dictionary filename
     */
    public String getDictionaryFilename() {
        return this.dictionaryFileName;
    }

    /**
     * Provides the name of this dictionary
     *
     * @return the string name of this dictionary
     */
    public String getDictionaryName() {
        return this.dictionaryName;
    }

    /**
     * Provides the unique identifier for this dictionary
     *
     * @return the string GUID for dictionary
     */
    public String getGUID() {
        return this.GUID;
    }

    /**
     * Provides the largest IDF value in this dictionary
     *
     * @return the <code>double</code> value of the largest IDF in this dictionary
     */
    public double getMaxIDFFound() {
        return this.maxIDFfound;
    }

    /**
     * Provides the largest IDF value in this dictionary
     *
     * @return the <code>double</code> value of the largest IDF in this dictionary
     */
    public double getMaxIDF() {
        return this.trimmed_max;
    }

    /**
     * Provides the largest IDF value in this dictionary
     *
     * @return the <code>double</code> value of the largest IDF in this dictionary
     */
    public double getMinIDF() {
        return this.trimmed_min;
    }

    /**
     * Provides the normalized inverse document frequency (IDF) of the given token in this dictionary. The normalized IDF
     * is the token's IDF divided by the highest IDF in this dictionary.
     *
     * @param token the string containing the token value
     * @return the <code>double</code> value of the token's normalized inverse document frequency (IDF); returns 0.0 if
     * the token is not present in this dictionary
     */
    public synchronized Double getNormalizedIDF(Token token) {

        if (this.dictionary.containsKey(token)) {
            return getNormalizedIDF(this.dictionary.get(token));
        }

        return 0.0;
    }

    /**
     * Provides the normalized inverse document frequency (IDF) of the given dictionaryEntry. The normalized IDF is the
     * token's IDF divided by the highest IDF in this dictionary.
     *
     * @param entry the dictionaryEntry to calculate the normalized IDF
     * @return the <code>double</code> value of the dictionaryEntry's normalized inverse document frequency (IDF)
     */
    public Double getNormalizedIDF(DictionaryEntry entry) {
        return entry.getIDF() / this.maxIDFfound;
    }

    /**
     * Provides the position of the given token in this dictionary
     *
     * @param token the string containing the token value
     * @return the position of the token; returns -1 if the token is not present in this dictionary
     */
    public int getPosition(Token token) {
        if (this.dictionary.containsKey(token)) {
            return this.dictionary.get(token).getPosition();
        }
        return -1;
    }

    /**
     * Provides the number of tokens in this dictionary
     *
     * @return the number of unique tokens in this dictionary
     */
    public int size() {
        return this.dictionary.size();

    }

    /**
     * Provides the directory used for this dictionary
     *
     * @return the string directory for which this dictionary was created
     */
    public String getSource() {
        return this.directory;
    }

    /**
     * Provides the identifier for the system that this dictionary was created on
     *
     * @return the string system ID for this dictionary
     */
    public String getSystemID() {
        return this.systemID;
    }

    /**
     * Provides the tokenizerList object used to create this dictionary
     *
     * @return the tokenizerList object used for this dictionary
     */
    public TokenizerList getTokenizers() {
        return this.tokenizers;
    }

    /**
     * Provides the names of the tokenizers
     *
     * @return the names of the tokenizers used for this dictionary
     */
    public List<String> getTokenizerNames() {
        return this.tokenizers.getNames();
    }


    /**
     * Provides a vector of the tokens in this dictionary
     *
     * @return the list of tokens in this dictionary
     */
    public synchronized List<Token> getTokens() {

        return Collections.synchronizedList(new Vector<>(this.dictionary.keySet()));

    }

    /**
     * Provides the number of documents processed for this dictionary
     *
     * @return the total number of documents used to create this dictionary
     */
    public int getTotalDocuments() {
        return this.totalDocuments;
    }

    /**
     * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs
     * dictionary to a file
     *
     * @param config_file the string name of the XML configuration file containing dictionary tokenizers
     * @param path        the string path to the directory used to create this dictionary
     * @param output_file the string name/path to write this dictionary to
     */
    public void makeDictionary(String config_file, String path, String output_file) throws SdTextException {

	
        try {	   

	    logger.info("Entering makeDictionary");
	    File config_path = FileChecker.canRead(config_file, "Dictionary", "makeDictionary");

	    File file_path = FileChecker.canRead(path, "Dictionary", "makeDictionary");

	    // If we have an output file path, make sure it works
	    if (output_file != null){
		File output_path = FileChecker.canWrite(output_file, "Dictionary", "makeDictionary");
	    }
	    
            addTokenizers(config_file);

            checkDictionaryException("makeDictionary 1");

            FileLister files = new FileLister(path);

            files.loadQueue();	    
	    
            createFullDictionary(files);

            //Global.debug("File size: ", files.size());

            if (output_file != null) {
                saveDictionaryXML(output_file);
            } else {
		stdOutDictionaryXML();
	    }
	    
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }

	logger.info("Leaving makeDictionary");
    }

    /**
     * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs a
     * trimmed dictionary to a file. Used by the GUI, not the command line version
     *
     * @param tokenizersNames the list of names of the dictionary tokenizers
     * @param path            the string path to the directory used to create this dictionary
     * @param min_idf         the <code>double</code> minimum normalized inverse document frequency (IDF) to trim dictionary
     * @param max_idf         the <code>double</code> maximum normalized inverse document frequency (IDF) to trim dictionary
     * @param output_file     the string name/path to write this dictionary to
     */
    public void makeDictionary(List<String> tokenizersNames,
                               String path,
                               double min_idf,
                               double max_idf,
                               String output_file) throws SdTextException {

        try {

	    File file_path = FileChecker.canRead(path, "Dictionary", "makeDictionary");

	    File output_path = FileChecker.canWrite(output_file, "Dictionary", "makeDictionary");
	    	   	

            setTokenizers(tokenizersNames);

            checkDictionaryException("makeDictionary 2");


            FileLister files = new FileLister(path);
	    files.loadQueue();	    
            createFullDictionary(files);

            if (min_idf > 0 || max_idf < 1.0) {
                trimByIDF(min_idf, max_idf);
            }
            saveDictionaryXML(output_file);
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    private void checkDictionaryException(String method) throws SdTextException {

        try {
            if (this.tokenizers.getNames() == null) {
		logger.warn("Null tokenizer list");
                ExceptionHandler.handleException(new Exception("Dictionary:" + method + ":Null tokenizerList"));
            }
            if (this.tokenizers.getNames().size() < 1) {
		logger.warn("Empty tokenizer list");
                ExceptionHandler.handleException(new Exception("Dictionary:" + method + ":Empty tokenizerList"));
            }
            if (this.dictionary == null) {
		logger.warn("Null dictionary");
                ExceptionHandler.handleException(new Exception("Dictionary:" + method + ":Null dictionary"));
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Merges the given Map of tokens to dictionaryEntry with this dictionary
     *
     * @param documentTokens the map containing the dictionary information to merge with this dictionary
     */
    protected synchronized void mergeDictionary(Map<Token, DictionaryEntry> documentTokens) {
        for (Entry<Token, DictionaryEntry> documentEntry : documentTokens.entrySet()) {
            if (this.dictionary.containsKey(documentEntry.getKey())) {
                // If so, increment the document count and frequency count
                this.dictionary.get(documentEntry.getKey()).incrementFrequencyCount(
                        documentEntry.getValue().getFrequencyCount());
                this.dictionary.get(documentEntry.getKey()).incrementDocumentCount();
            } else {
                // Otherwise, insert the DictionaryEntry into this dictionary
                documentEntry.getValue().setPosition(this.nextPosition);
                this.nextPosition++;
                this.dictionary.put(documentEntry.getKey(), documentEntry.getValue());
            }
        }
        this.totalDocuments++;
    }

    /**
     * Writes this dictionary as a serialized XML object.
     *
     * @param filename the string XML filename to write this dictionary to
     */
    public void saveDictionaryXML(String filename) throws SdTextException {

        if(dictionary.isEmpty()) {
            throw new SdTextException("Dictonary:saveDictionaryXML:No tokens found for dictionary. This happens if there is insufficient variety in files used to create the dictionary, or if it is trimmed too tightly.");
        }

        try {

            this.dictionaryFileName = (new File(filename)).getAbsolutePath();
            this.dictionaryName = filename.substring(filename.lastIndexOf('/') + 1);
            this.creation = String.valueOf(new Date());
            DecimalFormat rounding = new DecimalFormat("#0.####");

            Element dict = new Element("dictionary");
            dict.setAttribute("name", this.dictionaryFileName);
            dict.setAttribute("GUID", this.GUID);
            dict.setAttribute("systemID", this.systemID);
            if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
                dict.setAttribute("creatingProgram", this.creatingProgram);
            if (getCreation() != null && getCreation().length() > 0)
                dict.setAttribute("creation", this.creation);
            if (getSource() != null && getSource().length() > 0)
                dict.setAttribute("directory", this.directory);
            dict.setAttribute("maxIDF", rounding.format((this.maxIDFfound)));
            dict.setAttribute("totalDocuments", String.valueOf(this.totalDocuments));
            dict.setAttribute("trimmed", String.valueOf(this.trimmed));
            if (this.trimmed) {
                dict.setAttribute("trimmed_min", String.valueOf(this.trimmed_min));
                dict.setAttribute("trimmed_max", String.valueOf(this.trimmed_max));
            }
            synchronized (this) {
                dict.setAttribute("nextPosition", String.valueOf(this.nextPosition));
            }
            for (String tokenizer : this.tokenizers.getNames()) {
                String[] split = tokenizer.split("\\s");
                Element tokenizerElement = new Element("tokenizer");
                tokenizerElement.setAttribute("name", split[0]);
                StringBuffer parameters = new StringBuffer();
                if (split.length > 1) {
                    parameters.append(split[1]);
                    tokenizerElement.setAttribute("parameter", split[1]);
                }
                dict.addContent(tokenizerElement);
            }

            for (Entry<Token, DictionaryEntry> entry : this) {
                try {
                    Element tokenElement = new Element("token");
                    tokenElement.setAttribute("name", entry.getKey().name());
                    tokenElement.setAttribute("start_byte", String.valueOf(entry.getKey().start_byte()));
                    tokenElement.setAttribute("end_byte", String.valueOf(entry.getKey().end_byte()));
                    tokenElement.setAttribute("position", String.valueOf(entry.getValue().getPosition()));
                    tokenElement.setAttribute("frequency", String.valueOf(entry.getValue().getFrequencyCount()));
                    tokenElement.setAttribute("documentCount", String.valueOf(entry.getValue().getDocumentCount()));
                    tokenElement.setAttribute("idf", rounding.format((entry.getValue().getIDF())));
                    tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(entry.getValue()))));
                    dict.addContent(tokenElement);
                } catch (IllegalDataException exception) {

                    logger.error("Dictionary.org.jdom.IllegalDataException prohibits token from " +
                            "being added to dictionary: " + entry.getKey());
                }
            }

            try {
                Document doc = new Document(dict);
                FileOutputStream dictionaryOut = new FileOutputStream(this.dictionaryFileName);
                Format form = Format.getPrettyFormat();
                form.setEncoding(Global.fileEncoding);
                XMLOutputter out = new XMLOutputter(form);
                out.output(doc, dictionaryOut);

                logger.info("Writing dictionary to {}", this.dictionaryFileName);

                dictionaryOut.close();
            } catch (IOException e) {
                ExceptionHandler.handleException(e);
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Writes dictionary to standard output
     */
    public void stdOutDictionaryXML() throws SdTextException {
        try {

            this.creation = String.valueOf(new Date());
            DecimalFormat rounding = new DecimalFormat("#0.####");

            Element dict = new Element("dictionary");
            //  dict.setAttribute("name", this.dictionaryFileName);
            dict.setAttribute("GUID", this.GUID);
            dict.setAttribute("systemID", this.systemID);
            if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
                dict.setAttribute("creatingProgram", this.creatingProgram);
            if (getCreation() != null && getCreation().length() > 0)
                dict.setAttribute("creation", this.creation);
            if (getSource() != null && getSource().length() > 0)
                dict.setAttribute("directory", this.directory);
            dict.setAttribute("maxIDF", rounding.format((this.maxIDFfound)));
            dict.setAttribute("totalDocuments", String.valueOf(this.totalDocuments));
            dict.setAttribute("trimmed", String.valueOf(this.trimmed));
            if (this.trimmed) {
                dict.setAttribute("trimmed_min", String.valueOf(this.trimmed_min));
                dict.setAttribute("trimmed_max", String.valueOf(this.trimmed_max));
            }
            synchronized (this) {
                dict.setAttribute("nextPosition", String.valueOf(this.nextPosition));
            }
            for (String tokenizer : this.tokenizers.getNames()) {
                String[] split = tokenizer.split("\\s");
                Element tokenizerElement = new Element("tokenizer");
                tokenizerElement.setAttribute("name", split[0]);
                if (split.length > 1) {
                    tokenizerElement.setAttribute("parameter", split[1]);
                }
                dict.addContent(tokenizerElement);
            }

            for (Entry<Token, DictionaryEntry> entry : this) {
                try {
                    Element tokenElement = new Element("token");
                    tokenElement.setAttribute("name", entry.getKey().name());
                    tokenElement.setAttribute("start_byte", String.valueOf(entry.getKey().start_byte()));
                    tokenElement.setAttribute("end_byte", String.valueOf(entry.getKey().end_byte()));
                    tokenElement.setAttribute("position", String.valueOf(entry.getValue().getPosition()));
                    tokenElement.setAttribute("frequency", String.valueOf(entry.getValue().getFrequencyCount()));
                    tokenElement.setAttribute("documentCount", String.valueOf(entry.getValue().getDocumentCount()));
                    tokenElement.setAttribute("idf", rounding.format((entry.getValue().getIDF())));
                    tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(entry.getValue()))));
                    dict.addContent(tokenElement);
                } catch (IllegalDataException exception) {

                    logger.error("org.jdom.IllegalDataException prohibits token" +
                            " from being added to dictionary: ", entry.getKey());
                }
            }

            try {
                Document doc = new Document(dict);
                Format form = Format.getPrettyFormat();
                form.setEncoding(Global.fileEncoding);
                XMLOutputter printer = new XMLOutputter(form);
                printer.output(doc, System.out);
                System.out.println("Writing dictionary to standard out");

            } catch (IOException e) {
                ExceptionHandler.handleException(e);
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Determines the name of the program that created this dictionary
     */
    private void setCreatingProgram() throws SdTextException {
        try {
            logger.info("Start");
            this.creatingProgram = "";
            Runtime runtime = Runtime.getRuntime();
            String cmds[] = {"jps", "-m"};
            Process proc;
            try {
                proc = runtime.exec(cmds);
                BufferedReader bufferedreader = new BufferedReader(new InputStreamReader(proc.getInputStream(), Global.fileEncoding));

                String line;
                while ((line = bufferedreader.readLine()) != null) {
                    for (String output : line.split("\\s")) {
                        if (output.contains(".jar")) {
                            this.creatingProgram = output;

                        }
                    }
                }
                bufferedreader.close();
                proc.getInputStream().close();
                proc.getOutputStream().flush();
                proc.getOutputStream().close();
                proc.getErrorStream().close();

            } catch (IOException e) {
                ExceptionHandler.handleException(e);
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Sets the name of the program that created this dictionary
     *
     * @param creatingProgram the string program that created this dictionary
     */
    public void setCreatingProgram(String creatingProgram) {
        this.creatingProgram = creatingProgram;

    }

    /**
     * Sets the system ID of this dictionary
     */
    private void setSystemID() throws SdTextException {
        try {
            this.systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
        } catch (UnknownHostException e) {
            ExceptionHandler.handleException(e);
        }

    }

    /**
     * Loads the tokenizers to use for this dictionary from a TokenizerList object. Only works if no tokenizers have
     * already been set.
     *
     * @param tokenizerList the TokenizerList object containing the tokenizers to use for this dictionary
     */
    public void setTokenizers(TokenizerList tokenizerList) throws SdTextException {
	if (tokenizerList == null){
	    ExceptionHandler.handleException(new Exception("Null tokenizer list"));
	}

	if (this.tokenizers == null) {
            this.tokenizers = tokenizerList;
        } else {
            ExceptionHandler.handleException(new Exception("Dictionary: Cannot reset tokenizer list in a dictionary"));
        }

    }

    /**
     * Sets the tokenizers used by this dictionary; Only works if no tokenizers have already been set or if the given list
     * contains the same tokenizers as those that have already been set
     *
     * @param tokenizerVec the list of tokenizer names to use to create this dictionary
     */
    public void setTokenizers(List<String> tokenizerVec) throws SdTextException {
        try {
            if (this.tokenizers == null) {
                this.tokenizers = new TokenizerList(tokenizerVec);
            } else if (this.tokenizers.getNames().size() != tokenizerVec.size()
                    || !tokenizerVec.containsAll(this.tokenizers.getNames())) {

                logger.info("Dictionary cannot reset tokenizers in a dictionary");
                ExceptionHandler.handleException(new Exception("Dictionary: Cannot reset tokenizers in a dictionary"));
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }

    }

    /**
     * Prints this dictionary statistics, including:
     * <ul>
     * <li>Name</li>
     * <li>Number of documents</li>
     * <li>Number of tokens</li>
     * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
     * <li>Minimum and maximum IDF</li>
     * <li>List of tokenizers</li>
     * </ul>
     */
    void showStatistics() {
        System.out.println("Name: " + this.dictionaryFileName);
        System.out.println("# Docs: " + this.totalDocuments);
        System.out.println("# Tokens: " + this.dictionary.size());
        if (this.trimmed) {
            System.out
                    .println("Trimmed: " + this.trimmed + " (min: " + this.trimmed_min + ", max: " + this.trimmed_max + ")");
        } else {
            System.out.println("Trimmed: " + this.trimmed);
        }
        System.out.println("Max IDF: " + this.maxIDFfound);
        System.out.println("Tokenizers: " + this.tokenizers);
    }

    /**
     * Provides this dictionary statistics, including:
     * <ul>
     * <li>Name</li>
     * <li>Number of documents</li>
     * <li>Number of tokens</li>
     * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
     * <li>Minimum and maximum IDF</li>
     * <li>List of tokenizers</li>
     * </ul>
     *
     * @return the string containing this dictionary's statistics
     */
    public String getStatistics() {
        StringBuilder statisticsBuffer = new StringBuilder();
        statisticsBuffer.append(String.format("%s%n", "Name: " + this.dictionaryFileName));
        statisticsBuffer.append(String.format("%s%n", "# Docs: " + this.totalDocuments));
        statisticsBuffer.append(String.format("%s%n", "# Tokens: " + this.dictionary.size()));

        if (this.trimmed) {
            statisticsBuffer.append(String.format("%s%n", "Trimmed: " + this.trimmed + " (min: " + this.trimmed_min
                    + ", max: " + this.trimmed_max + ")"));
        } else {
            statisticsBuffer.append(String.format("%s%n", "Trimmed: " + this.trimmed));
        }
        statisticsBuffer.append(String.format("%s%n", "Max IDF: " + this.maxIDFfound));
        statisticsBuffer.append(String.format("%s%n", "Tokenizers: " + this.tokenizers));
        return statisticsBuffer.toString();
    }

    /**
     * Prints each token in this dictionary with it's frequency, IDF, and normalized IDF
     */

    void showTokens() {
        System.out.format("%-25s %-10s %-10s %-10s%n", "Token", "Freq", "IDF", "Norm. IDF");
        System.out.println("----------------------------------------------------------");
        for (Entry<Token, DictionaryEntry> entry : this) {
            System.out.format("%-25s %-10s %-10.4f %-10.4f%n", entry.getKey(), entry.getValue().getFrequencyCount(), entry
                    .getValue().getIDF(), getNormalizedIDF(entry.getValue()));
        }
    }

    /**
     * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
     * inside a threaded application).
     *
     * @param numThreads the number of threads to use to create this dictionary
     */
    public void threadingOn(int numThreads) {
        this.maxThread = numThreads;

    }

    /**
     * Trims this dictionary by removing any token that is outside a range of normalized IDFs. Once trimmed, no new tokens
     * can be added to this dictionary.
     *
     * @param min the <code>double</code> minimum normalized IDF to keep; any token with a lower IDF will be discarded.
     * @param max the <code>double</code> maximum normalized IDF to keep; any token with a higher IDF will be discarded.
     */
    public void trimByIDF(double min, double max) throws SdTextException {

        try {
            logger.info("Trimming the dictionary");
	    
	    if (min < 0 || min > 1 || max < 0 || max > 1){
		ExceptionHandler.handleException(new Exception("IDF must be in the range 0 to 1 inclusive"));
	    }
	    
            if (min >= max) {
                logger.info("IDF Range Max {} cannot be greater than min {}", max,  min);
                ExceptionHandler.handleException(new Exception("IDF Range cannot be greater than min"));
            }

            // Trims this dictionary to contain only tokens within the specified range of normalized IDF values
            if (this.trimmed) {
                logger.info("Dictionary trimming an already trimmed dictionary might be a bad idea");

            }

            this.trimmed = true;
            this.trimmed_min = min;
            this.trimmed_max = max;


            int newposition = 0;
            HashMap<Token, DictionaryEntry> newdict = new HashMap<>();

            Iterator<Entry<Token, DictionaryEntry>> iterator = this.dictionary.entrySet().iterator();

            while (iterator.hasNext()) {
                Entry<Token, DictionaryEntry> entry = iterator.next();
		Double currentIDF = getNormalizedIDF(entry.getValue());
		if (!currentIDF.isNaN()){
			if ((currentIDF >= min) && (currentIDF <= max)) {			    
			    entry.getValue().setPosition(newposition);
			    newposition++;
			    newdict.put(entry.getKey(), entry.getValue());
			}
		    }

            }
            this.dictionary = newdict;
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        logger.info("end");
    }


    /**
     * @return a dictionary's tokens with their IDF ranges
     */
    public Set<Pair<Token, Double>> getDictionary() {

        Set<Pair<Token, Double>> tempSet = new TreeSet<>();

        for (Map.Entry<Token, DictionaryEntry> p : dictionary.entrySet()) {

            tempSet.add(new Pair<>(p.getKey(), p.getValue().getIDF()));

        }

        return tempSet;

    }

    /**
     * @return the dictionary positions
     */
    public Map<Integer, Token> getDictionaryAndPosition() {

        Map<Integer, Token> tempSet = new HashMap<>();

        for (Map.Entry<Token, DictionaryEntry> p : dictionary.entrySet()) {

            tempSet.put(p.getValue().getPosition(), p.getKey());

        }

        return tempSet;

    }

    public Map<Token, DictionaryEntry> getFullDictionary() {
        return dictionary;
    }

}
