package edu.georgetown.gucs.dictionary;

import java.io.File;
import edu.georgetown.gucs.utility.*;


/**
 * Creates a {@linkplain Dictionary} containing tokens with IDFs within a specified range
 *
 * @author Clay Shields
 */
public class TrimDictionary {

    // removed final may be problematic
    private  Dictionary dictionary  ;
    private double min_idf = 0.0;
    private double max_idf = 1.0;

    /**
     * Constructor that specifies the IDF range and dictionary to trim
     *
     * @param min_idf         the <code>double</code> minimum normalized inverse document frequency (IDF) to use for this trimmed
     *                        dictionary
     * @param max_idf         the <code>double</code> maximum normalized inverse document frequency (IDF) to use for this trimmed
     *                        dictionary
     * @param dictionary_file the string filename containing the dictionary to trim
     */
    public TrimDictionary(double min_idf, double max_idf, String dictionary_file) throws SdTextException {

        try {
	    File dictionaryFile = FileChecker.canRead(dictionary_file, "TrimDictionary", "TrimDictionary");

	    if ((min_idf < 0) || (min_idf > 1)){
		throw new SdTextException("TrimDictionary:TrimDictionary:min IDF range is 0 to 1");
	    }

	    if ((max_idf < 0) || (max_idf > 1)){
		throw new SdTextException("TrimDictionary:TrimDictionary:max IDF range is 0 to 1");
	    }

	    if (min_idf >= max_idf){
		throw new SdTextException("TrimDictionary:TrimDictionary:min IDF must be less than max IDF");
	    }
		  	    
	    this.dictionary = new Dictionary(dictionary_file);
            this.min_idf = min_idf;
            this.max_idf = max_idf;
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Trims the dictionary and saves it to the given output file
     *
     * @param output_file the string filename to write the trimmed dictionary to
     */
    public void trim(String output_file) throws SdTextException {
        try {
            this.dictionary.trimByIDF(this.min_idf, this.max_idf);
            this.dictionary.saveDictionaryXML(output_file);
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

}
