package edu.georgetown.gucs.experiment;

import edu.georgetown.gucs.tokenizers.TokenizerParameters;
import edu.georgetown.gucs.utility.*;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.file.FileSystems;
import java.nio.file.Path;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Evan B.
 * @version 2.0
 * @since 06/16/14
 */
public class ExperimentConfReader {

    private String configFile;
    private String dataSet;
    private Path path;
    private String plotsPath;
    private String description;
    private String manglerPlots;
    private String language;

    private Set<String> plots = new HashSet<>();
    ;
    private List<String> fingerprinters = new Vector<>();

    private List<Long> randomSeeds = new Vector<>();


    private List<Double> sampleParameters = new Vector<>();
    private List<Double> dictionaryParameters = new Vector<>();

    private Vector<Pair<Double, Double>> IDFs = new Vector<>();
    private Vector<Pair<String, Integer>> matcherVector = new Vector<>();


    private List<TokenizerParameters> tokenizerParametersList = new Vector<>();

    private int numberOfTrials;
    private int numberOfTrialThreads;

    private File datasetPath;


    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(ExperimentConfReader.class);


    /**
     * Default constructor
     * sets the configFile and loads the config file
     *
     * @param configFile sets parameter information
     */
    ExperimentConfReader(String configFile) throws SdTextException {

        try {

	    FileChecker.canRead(configFile, "ExperimentConfReader", "ExperimentConfReader"); 
	    
	    System.err.println("In ExperimentConfReader constructor");
	    
	    this.configFile = configFile;

            logger.info("[loading xml config]");
	    System.err.println("ExperimentConfReader about to call loadXMLconfig");
            loadXMLConfig();
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }


    /**
     * Reads a config file to parse the following:
     * [Number of trials]
     * [description]
     * [plots]
     * --> finish
     */
    private void loadXMLConfig() throws SdTextException {

        try {
            File xml = new File(configFile);
            SAXBuilder builder = new SAXBuilder();

            try {				

		Document doc = builder.build(xml);

                Element root = doc.getRootElement();

                String threads = root.getAttributeValue("threads");

		
                numberOfTrialThreads = (threads != null && threads.length() > 0) ?
                        Integer.valueOf(threads) : Runtime.getRuntime().availableProcessors();

                String trials = root.getAttributeValue("trials");

                numberOfTrials = (trials == null || Integer.valueOf(trials) < 3) ? 3 :
                        Integer.valueOf(trials);
                logger.info("Number of trials {} ", numberOfTrials);

                if ((description = root.getAttributeValue("description")) == null) {

                    logger.error("Experiment: WARNING: Null description");
                    logger.info("Warning Null Description");
                }


                if (root.indexOf(root.getChild("plots")) != -1) {

                    plotsPath = root.getChild("plots").getAttributeValue("path");


                    @SuppressWarnings("unchecked")
                    List<Element> e = (List<Element>) root.getChild("plots").getChildren("plot");

                    for (Element plot : e) {

                        plots.add(plot.getAttributeValue("name")
                                .trim().toLowerCase());

                        if (plot.getAttributeValue("name").equalsIgnoreCase("mangler")) {

                            manglerPlots += (plot.getAttributeValue("parameter") == null) ?
                                    "" : plot.getAttributeValue("parameter")
                                    .toLowerCase(Locale.ENGLISH) + " ";
                        }
                    }
                }


                @SuppressWarnings("unchecked")
                List<Element> randSeed = root.getChildren("randomSeed");

                for (Element rand : randSeed) {

                    randomSeeds.add(Long.valueOf(rand.getAttributeValue("seed")));

                }

                if (randomSeeds.size() != 0 && numberOfTrials != randomSeeds.size()) {

                    logger.info("Experiment: Number of trials does not match number of random seeds. "
                            + "Trials: "
                            + this.numberOfTrials
                            + " seeds size: " + randomSeeds.size());

                    ExceptionHandler.handleException(new Exception(
                            "Experiment: Number of trials does not match number of random seeds. "
                                    + "Trials: "
                                    + this.numberOfTrials
                                    + " seeds size: " + randomSeeds.size()));
                }


		
                Element data = root.getChild("dataset");
                dataSet = data.getAttributeValue("name");

                if (dataSet == null || dataSet.length() == 0) {

                    ExceptionHandler.handleException(
                            new Exception("Experiment: No dataset name provided"));
                    logger.info("Experiment: No dataset name provided");

                }
                logger.info( "Dataset: {}", dataSet);

                String path = data.getAttributeValue("path");
                if (path != null) {

                    datasetPath = new File(path);
                    this.path = FileSystems.getDefault().getPath(path);
                } else {

                    logger.error("No path to dataset provided");
                    ExceptionHandler.handleException(
                            new Exception("Experiment: No path to dataset provided"));
                }
                logger.error("datasetPath {}", datasetPath);

                @SuppressWarnings("unchecked")
                List<Element> sampleElem = data.getChildren("sample");

                for (Element sampleParameter : sampleElem) {

                    if (sampleParameter.getAttributeValue("count") != null) {

                        double parameter = Double.valueOf(sampleParameter.getAttributeValue("count"));
                        logger.info("[SampleParameter attribute] "
                                + Double.valueOf(sampleParameter.getAttributeValue("count")));

                        if (parameter < 0) {
                            ExceptionHandler.handleException(new Exception("Experiment: Negative sample counts"));
                            logger.info("Experiment: Negative sample counts]");

                        } else if (parameter > 5000) {

                            logger.error( "LoadXMLConfig: Sample counts greater than " +
                                    "5000 can take a really long time when run with many manglers");
                        }
                        sampleParameters.add(parameter);
                    }

                    if (sampleParameter.getAttributeValue("percent") != null) {
                        double parameter = Double.valueOf(sampleParameter.getAttributeValue("percent"));

                        if (parameter < 1 || parameter > 100) {

                            logger.error("Sample Percentage out of range");

                            ExceptionHandler.handleException(
                                    new Exception("Experiment: Sample percentage out of range"));
                        }
                        this.sampleParameters.add(parameter / 1000);
                    }
                }
                if (this.sampleParameters == null || this.sampleParameters.size() == 0) {
                    logger.error("No sample settings provided");
                    ExceptionHandler.handleException(new Exception("Experiment: No sample settings provided"));
                }

                Element dict = data.getChild("dictionary");
                String lang = dict.getAttributeValue("language");
                language = (lang != null) ? lang : "english";

                @SuppressWarnings("unchecked")
                List<Element> create = dict.getChildren("create");

                for (Element dictParameter : create) {

                    if (dictParameter.getAttributeValue("count") != null) {
                        double parameter = Double.valueOf(dictParameter.getAttributeValue("count"));

                        if (parameter < 0) {

                            logger.info("Negative dictionary counts");
                            ExceptionHandler.handleException(
                                    new Exception("Experiment: Negative dictionary counts"));
                        } else if (parameter == 0) {

                            logger.info("Zero dictionary counts");
                            ExceptionHandler.handleException(
                                    new Exception("Experiment: Zero dictionary counts"));

                        }
                        logger.info("[Dict File Size] " + parameter);
                        dictionaryParameters.add(parameter);
                        logger.info("dictionaryParams size " + dictionaryParameters.size());

                    }

                    if (dictParameter.getAttributeValue("percent") != null) {
                        double parameter = Double.valueOf(dictParameter.getAttributeValue("percent"));

                        if (parameter < 1 || parameter > 100) {
                            logger.error("Percentage out of range");
                            ExceptionHandler.handleException(new Exception("Experiment: Dictionary percentage out of range"));
                        }
                        this.dictionaryParameters.add(parameter / 1000);
                        logger.info("[Percent dictionary count] " + parameter / 1000);
                    }
                }

                if (this.dictionaryParameters == null || this.dictionaryParameters.size() == 0) {
                    logger.error("No dictionary settings provided");
                    ExceptionHandler.handleException(
                            new Exception("Experiment: No dictionary settings provided"));
                }

                if (dict.getChild("IDF") != null) {

                    @SuppressWarnings("unchecked")
                    List<Element> dictChild = dict.getChildren("IDF");

                    for (Element idf : dictChild) {

                        Double minIDF = Double.valueOf(idf.getAttributeValue("min"));
                        Double maxIDF = Double.valueOf(idf.getAttributeValue("max"));

                        if (minIDF < maxIDF) {

                            IDFs.add(new Pair<>(minIDF, maxIDF));

                        } else {

                            logger.error("IDF out of range");
                            ExceptionHandler.handleException(
                                    new Exception("Experiment: The minimum IDF value (" + minIDF
                                            + ") is >= the maximum IDF value (" + maxIDF + ")"));
                        }
                    }

                } else {
                    IDFs.add(new Pair<>(0.0, 1.0));
                }

                @SuppressWarnings("unchecked")
                List<Element> children = root.getChildren("fingerprinter");


                for (Element fingerprint : children) {

                    fingerprinters.add(fingerprint.getAttributeValue("name"));

                }

                logger.info("[Fingerprinters] " + fingerprinters);

                if (fingerprinters == null || fingerprinters.size() == 0) {
                    logger.error("No Fingerprinter provided");

                    ExceptionHandler.handleException(
                            new Exception("Experiment: No fingerprinter provided"));
                }

                @SuppressWarnings("unchecked")
                List<Element> matcher = root.getChildren("matcher");

                for (Element matcherElement : matcher) {
                    if (matcherElement.getAttributeValue("score") != null) {
                        this.matcherVector.add(new Pair<>(matcherElement.getAttributeValue("name"), Integer
                                .valueOf(matcherElement.getAttributeValue("score"))));
                    } else {
                        // If there are no matcher parameters then add the default minimum score (0) to the vector
                        this.matcherVector.add(new Pair<>(matcherElement.getAttributeValue("name"), 0));
                    }
                }
                if (this.matcherVector == null || this.matcherVector.isEmpty()) {
                    logger.error("No matcher provided");
                    ExceptionHandler.handleException(new Exception("Experiment: No matcher provided"));
                }

                loadTokenizers(root);


            } catch (FileNotFoundException e) {
                logger.error("File Not found");

                ExceptionHandler.handleException(e);
            } catch (UnsupportedEncodingException e) {

                logger.error("Unsuported Encoding Exception");
                ExceptionHandler.handleException(e);

            } catch (JDOMException e) {

                logger.error("JDomException");
                ExceptionHandler.handleException(e);

            } catch (IOException e) {

                logger.error("IOException");
                ExceptionHandler.handleException(e);
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    private void loadTokenizers(Element root) throws SdTextException {
        try {
            @SuppressWarnings("unchecked")
            List<Element> tokenizerChildren = root.getChildren("tokenizers");
            for (Element element : tokenizerChildren) {
                Vector<String> tokenizers = new Vector<>();
                Vector<String> manglers = new Vector<>();
                Vector<String> splitters = new Vector<>();

                @SuppressWarnings("unchecked")
                List<Element> tokenizerList = element.getChildren("tokenizer");
                for (Element tokenizerElement : tokenizerList) {

                    String tokenizer = tokenizerElement.getAttributeValue("name");
                    if (tokenizers.size() == 0 && !tokenizer.contains("FileTokenizer")) {
                        ExceptionHandler.handleException(new Exception("Experiment: First tokenizer (" + tokenizer
                                + ") must read from a file."));
                    } else {

                        @SuppressWarnings("unchecked")
                        List<Element> splitterList = tokenizerElement.getChildren("splitter");

                        for (Element splitterElement : splitterList) {
                            splitters.add(splitterElement.getAttributeValue("name")
                                    + " " + splitterElement.getAttributeValue("value"));
                        }
                    }

                    if (tokenizerElement.getAttributeValue("parameter") != null) {
                        tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
                    } else if (tokenizer.equalsIgnoreCase("FileManglerTokenizer")) {

                        @SuppressWarnings("unchecked")
                        List<Element> tokenizerGetChildren = tokenizerElement.getChildren("manglers");

                        for (Element manglersListElement : tokenizerGetChildren) {


                            @SuppressWarnings("unchecked")
                            List<Element> manglerXML = manglersListElement.getChildren("mangler");

                            //changed from stringBuffer (synchronized)
                            StringBuilder mangler = new StringBuilder();
                            for (Element manglerElement : manglerXML) {

                                mangler.append(" ").append(manglerElement.getAttributeValue("name").trim());

                                if (!manglerElement.getAttributeValue("name").equalsIgnoreCase("null")) {
                                    mangler.append("-").append(manglerElement.getAttributeValue("percent").trim());
                                }
                            }
                            manglers.add(mangler.toString().trim());
                        }
                        if (manglers.size() == 0) {
                            ExceptionHandler.handleException(
                                    new Exception("Experiment: No manglers provided"));
                        }
                    }
                    tokenizers.add(tokenizer.trim());
                }

                if (splitters.size() == 0) {
                    splitters.add("null");
                }

                this.tokenizerParametersList.add(
                        new TokenizerParameters(tokenizers, manglers, splitters));
            }
            if ((this.tokenizerParametersList.size() == 0)) {
                ExceptionHandler.handleException(
                        new Exception("Experiment: No tokenizers provided"));
            }
        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    public List<Double> getDictionaryParameters() {
        return dictionaryParameters;
    }

    public List<Double> getSampleParameters() {
        return sampleParameters;
    }


    public List<String> getFingerprinter() {
        return fingerprinters;
    }

    public List<TokenizerParameters> getTokenizerParameters() {
        return tokenizerParametersList;
    }

    public Vector<Pair<Double, Double>> getIDF() {
        return IDFs;
    }

    public List<Pair<String, Integer>> getMatcherVector() {
        return matcherVector;
    }

    public int getNumTrialThreads() {
        return numberOfTrialThreads;
    }

    public int getNumTrials() {
        return numberOfTrials;
    }

    public File getDataSetPath() {
        return datasetPath;
    }

    public Path getPath() {
        return path;
    }


    public String getDataSetDirectory() {
        return dataSet;
    }

    public int getTotalTrialCount() {

        ArrayList<Integer> paramTest = new ArrayList<>();
        for (TokenizerParameters params : tokenizerParametersList) {

            //multiply by splitter?
            paramTest.add(params.manglers().size());
            paramTest.add(params.splitters().size());

            /* //Debug Code
            System.out.println(params.manglers().size());
            System.out.println(params.tokenizers().size());
            System.out.println(params.splitters().size());
            */
        }

        int counter = 0;
        for (int test : paramTest) {
            // [Debug Code]
            // System.out.println("[MangTokeSplit] " + test);
            counter += test;
        }

        /*  //Debug Code
        System.out.println(getDictionaryParameters().size() + " "
                + " " + getSampleParameters().size()
                + " " + getFingerprinter().size()
                + " " + getIDF().size()
                + " " + getMatcherVector().size()
                + " " + getNumTrials()
                + " " + counter);

          int totalExperiments = this.numberOfTrials * this.dictionaryParameters.size() * this.sampleParameters.size()
	    * this.IDFs.size() * this.tokenizerParametersList.size() * manglerCount * splitterCount
	    * this.fingerprinters.size() * this.matcherVector.size();

        */

        return (getDictionaryParameters().size()
                * getSampleParameters().size()
                * getFingerprinter().size()
                * getIDF().size()
                * getMatcherVector().size()
                * getNumTrials()
                * tokenizerParametersList.size());

    }

}



