
package edu.georgetown.gucs.matcher;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;


import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.fingerprinter.BulkFingerprinter;
import edu.georgetown.gucs.utility.*;
import org.slf4j.LoggerFactory;


/**
 * Used for doing a comparison of all files in a directory 
 *
 *
 * @author Clay Shields
 */
public class CompareDirectory {


    protected FileLister fileLister;

    protected Fingerprinter fingerprint;
    protected FingerprintMatcher matcher;
    protected String matcherName;
    protected int minimum_score;
    protected Dictionary dictionary;
    protected SortedMap<Integer, List<Pair<String,String>>> scores;
    protected int maxThread;
    protected Fingerprint fing;

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(CompareDirectory.class);

    /**
     * Comparator that orders scores highest to lowest
     */
    private static class HightoLowScore implements Comparator<Integer>, Serializable {
        private static final long serialVersionUID = 5679507537454070748L;

        public HightoLowScore() {
        }

        @Override
        public int compare(Integer one, Integer two) {
            return two.compareTo(one);
        }
    }

    /**
     * Constructor that sets the matcher, fingerprint file, directory, minimum score and dictionary to use for this
     * comparison
     *
     * @param match           the string name of the matcher to use for this comparison
     * @param fingerprintFile the string filename of the fingerprint to compare with this directory
     * @param directory       the string directory path to compare with this fingerprint
     * @param minimumScore    the <code>int</code> minimum score needed to output a file for this comparison
     * @param dictionary      the string filename of the dictionary to use for this comparison; this dictionary will only be used if the
     *                        fingerprint file does not already contain a full dictionary
     *                        added global variables to set this.minimum_Score -- evan
     */
    public CompareDirectory(String match, String directory, int minimumScore, String dictionary) throws SdTextException {
        try {

	    //System.err.println("Entering CompareDirectory");
	    //System.err.println("CopreDirectory: minScore is: " + minimumScore);
	    
	    if (match == null){
		throw new SdTextException("CompareDirectory:CompareDirectory:null match");
	    }

	    if (minimumScore < Global.MIN_SCORE  || minimumScore > Global.MAX_SCORE){
		throw new SdTextException("CompareDirectory:CompareDirectory: minimum score out of range");
	    }
	    else
		this.minimum_score = minimumScore;

	    //System.err.println("Checking directory");

	    FileChecker.canRead(directory,"CompareDirectory","CompareDirectory");
	    FileChecker.canRead(dictionary,"CompareDirectory","CompareDirectory");
	    
	    //System.err.println("Building matcher");
	    
	    this.matcherName = match;
            this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);

	    //System.err.println("Building scores");
            this.scores = Collections.synchronizedSortedMap(new TreeMap<Integer, List<Pair<String,String>>>(new HightoLowScore()));

	    //System.err.println("Calling file lister");
            this.fileLister = new FileLister(directory);
            this.fileLister.loadQueue();

	    //System.err.println("Get the dictionary");
	    this.dictionary = new Dictionary(dictionary);

	    //System.err.println("Getting thread count");
            this.maxThread = Runtime.getRuntime().availableProcessors();
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }


    /**
     * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
     * matcher returns boolean values, true values are given a score of 100 and false values are given a score of 0
     */
    //HERE
    public List<Pair<Pair<String, String>, Integer>> compareDirectory() throws SdTextException {

	List<Pair<Pair<String, String>, Integer>> unsortedScores;
	
        try {

            List<Fingerprint> results = new ArrayList<>();

            BulkFingerprinter f = new BulkFingerprinter(dictionary, "BitVectorFingerprinter");

            //System.err.println("Fingerprinting directory... " + System.currentTimeMillis());
            results.addAll(f.computeAllFingerprints("BitVectorFingerprinter", dictionary, fileLister));
	    //System.err.println(" ...done " + System.currentTimeMillis());
	    
	    //System.err.println("Matching fingerprints... " + System.currentTimeMillis());

            BulkMatcher bulkMatch = new BulkMatcher(matcherName, minimum_score, maxThread);
	    unsortedScores = bulkMatch.compareAllFingerprints(results);

	    // Add the returned scores to the local sorted scores variable for printing and such
	    for (Pair<Pair<String, String>, Integer> temp : unsortedScores){

		Pair<String, String> currentFilenames = temp.getA();
		Integer currentScore = temp.getB();


		if (currentScore >= minimum_score){
		    List<Pair<String,String>> filesForScore = scores.get(currentScore);
		    
		    if (filesForScore == null){
			filesForScore = new ArrayList<Pair<String,String>>();
			filesForScore.add(currentFilenames);
			scores.put(currentScore, filesForScore);
		    }
		    else {		    
			filesForScore.add(currentFilenames);
		    }
		}
		else {
		    ///logger.
		}

	    }
	    
	    
	    //System.err.println(" ...done " + System.currentTimeMillis());

            return unsortedScores;

	} catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return null;
    }


    /**
     * Provides all the files and their similarity scores; if a minimum score is provided, only output those files with
     * scores greater than or equal to the minimum
     *
     * @return the string containing all the score results
     */  
    private String scoresToString() throws SdTextException {
        StringBuilder buffer = new StringBuilder();
        boolean printedSomething = false;
        try {
            for (int i = Global.MAX_SCORE; i >= this.minimum_score; i--) {
                if (this.scores.containsKey(i)) {
		    for (Pair<String, String> match : this.scores.get(i)) {
                        if (i < Global.MAX_SCORE && i > 10) buffer.append(String.format("%s%n", match.getA() + "|" + match.getB() + "|0" + i));
                        else if (i <= 10) buffer.append(String.format("%s%n", match.getA() + "|" + match.getB() + "|00" + i));
                        else buffer.append(String.format("%s%n", match.getA() + "|" + match.getB() + "|" + i));
                    }
                    printedSomething = true;
                }
            }
            if (!printedSomething) {
                buffer.append(String.format("%s",
                        "CompareDictionary: There are no matches with a score higher than " + this.minimum_score));
            }
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
        return buffer.toString();
    }

    /**
     * Print all the files and their similarity scores; if a minimum score is provided, only output those files with
     * scores greater than or equal to the minimum
     */
    public void printScores() throws SdTextException {
        try {
            System.out.println(scoresToString());
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

}
