package sdtext;

import argparser.*;
import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.dictionary.ShowDictionaryStatistics;
import edu.georgetown.gucs.dictionary.ShowDictionaryTokens;
import edu.georgetown.gucs.dictionary.TrimDictionary;
import edu.georgetown.gucs.experiment.Experiment;
import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.fingerprinter.ExtractFingerprintDictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.fingerprinter.SdhashFingerprinter;
import edu.georgetown.gucs.matcher.CompareFingerprint;
import edu.georgetown.gucs.matcher.CompareDirectory;
//import edu.georgetown.gucs.matcher.ExperimentN2;
import edu.georgetown.gucs.matcher.ScoreFingerprints;
import edu.georgetown.gucs.tokenizers.*;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.SdTextException;
import enums.CommandLineHelp;
import org.apache.log4j.LogManager;
import org.apache.log4j.PropertyConfigurator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Properties;

/**
 * @author Evan B.
 * @since 3/4/15
 * <p>
 * opensource.apple.com/source/Libc/Libc-320/include/sysexits.h
 */
public class CLIOptions extends Arguments {

    protected String programName_ = "";

    private static final Logger logger = LoggerFactory.getLogger(CLIOptions.class);
    private static final String TOKENIZER_CONFIG = "../doc/configuration_files/tokenizers_config.xml";

    public CLIOptions() throws SdTextException {
	throw new SdTextException("CLIOptions:CLIOptions:Call to default constructor. Arguments required");
    }

    /**
     * Retrieves the first option, the sdtext program to run, and passes the
     *
     * @param args the options passed from the commandline to SdText
     */
    public CLIOptions(String[] args) throws SdTextException {

	Global.setLoggingLvl("OFF");
	
	try {
	   	
	    ArgParser parser = new ArgParser("SdText CLI Parser");
	    
	    addOptions(parser);
	    
	    String[] getProgram = parser.matchAllArgs(args, 0, 0);	
	    
	    if (getProgram != null) {
		this.programName_ = getProgram[0].toUpperCase();
	    } //else throw help

	    //System.err.println ("Got log level of " + debug.value);			       
	    // Try to change the logging settings
	    if (debug.value != null){
		//System.err.println("Changing debug value" + debug.value);	    
		Global.setLoggingLvl(debug.value);
	    }
	    
	    
	}catch (SdTextException sdtext) {
	    throw sdtext;
	}
	
    }

    /**
     * Runs an SdText program with its parameters passed from the commandline
     */
    Boolean runProgram() throws SdTextException {

        //case with {}

        switch (programName_) {


			      
	    
	case "TOKENIZEFILE": {

	    logger.info("Running tokenizeFile");

	    if (input.value == null) printError(CommandLineHelp.TOKENIZEFILE);
	    if (config.value == null) {
		logger.info("Setting default tokenizers");
		config.value = TOKENIZER_CONFIG;
	    }
		
	    TokenizeFile tokenizeFile = new TokenizeFile(config.value);
	    tokenizeFile.tokenizeFile(input.value);
	    tokenizeFile.printTokens();

	    logger.info("Finishing tokenizeFile");
	    
	    return true;
	}

	case "ARABICFILETOKENIZER": {

	    logger.info("Running ArabicFileTokenizer");
	    
	    if (input.value == null) printError(CommandLineHelp.ARABICFILETOKENIZER);

	    ArabicFileTokenizer arabicTokenizer = new ArabicFileTokenizer();
	    arabicTokenizer.tokenize(input.value);
	    arabicTokenizer.printTokens();
	    logger.info("Finishing ArabicFileTokenizer");
	    return true;
	}

	case "CHINESEFILETOKENIZER": {

	    logger.info("Running ChineseFileTokenizer");
	    
	    if (input.value == null) printError(CommandLineHelp.CHINESEFILETOKENIZER);

	    ChineseFileTokenizer chineseTokenizer = new ChineseFileTokenizer();
	    System.out.println("Individual mode:");
	    chineseTokenizer.setMode("individual");
	    chineseTokenizer.tokenize(input.value);
	    chineseTokenizer.printTokens();
	    System.out.println("Smart mode:");
	    chineseTokenizer.setMode("smart");
	    chineseTokenizer.tokenize(input.value);
	    chineseTokenizer.printTokens();

	    logger.info("Finishing ChineseFileTokenizer");
	    
	    return true;
	}

	case "CREATEDICTIONARY": {


	    logger.info("Calling CreateDictionary");
	    
	    //Catch null values early to print help if needed
	    if (directory.value == null) {
		printError(CommandLineHelp.CREATEDICTIONARY);
	    }

	    if (config.value == null) config.value = TOKENIZER_CONFIG;

	    // Arguments are checked in makeDictionary, so if they are not null pass
	    // them through.
	    
	    Dictionary dict = new Dictionary();
	    logger.info("Calling makeDictionary with " + config.value + " "
			       + directory.value + " " + dictionary.value) ;
	    dict.makeDictionary(config.value, directory.value, dictionary.value);
	    logger.info("Leaving CreateDictionary");
	    return true;
	}

	case "SHOWDICTIONARYSTATISTICS":
	    if (dictionary.value == null) {
		printError(CommandLineHelp.SHOWDICTIONARYSTATISTICS);
	    }
	    ShowDictionaryStatistics.print(dictionary.value);
	    return true;

	case "SHOWDICTIONARYTOKENS": {

	    if (dictionary.value == null) {
		printError(CommandLineHelp.SHOWDICTIONARYTOKENS);
	    }
	    ShowDictionaryTokens.print(dictionary.value);
	    return true;
	}

	case "TRIMDICTIONARY": {

	    if (dictionary.value == null){
		printError(CommandLineHelp.TRIMDICTIONARY);		    
	    }

	    if (minIDF.value < 0 || minIDF.value > 1 ||
		maxIDF.value < 0 || maxIDF.value > 1){
		System.err.println("Bad IDF value l:" + minIDF.value + " h:" + maxIDF.value);
		printError(CommandLineHelp.TRIMDICTIONARY);		    
	    }

	    if (minIDF.value >= maxIDF.value) {
		System.err.println("Bad IDF range l:" + minIDF.value + " h:" + maxIDF.value);
		printError(CommandLineHelp.TRIMDICTIONARY);		    
                    
	    }
	    TrimDictionary trimDictionary = new TrimDictionary(minIDF.value, maxIDF.value, dictionary.value);
	    trimDictionary.trim(output.value);
	    return true;
	}
	    
	case "BITVECTORFINGERPRINTER": {

	    if (dictionary.value == null){
		System.err.println("Null dictionary");
		printError(CommandLineHelp.BITVECTORFINGERPRINTER);
	    }

	    if (input.value == null){
		System.err.println("Null input");
		printError(CommandLineHelp.BITVECTORFINGERPRINTER);
	    }

	    if (output.value == null) {		    
		System.err.println("Writing to fingerprint.xml");
		output.value = "fingerprint.xml";
	    }

		
	    if(config.value == null)  config.value = "../doc/configuration_files/fingerprinter_config.xml";

	    //System.err.println("Creating fingerprinter");
	    BitVectorFingerprinter bvfp = new BitVectorFingerprinter(dictionary.value);
	    //	    System.err.println("Creating adding config");
	    bvfp.outputFields(config.value);
	    //	    System.err.println("Computing fingerprint");
	    bvfp.generateXML(bvfp.computeFingerprint(input.value), input.value, output.value);
	    //	    System.err.println("Computing fingerprint");

	    return true;
	}
	    
	case "SCOREFINGERPRINTS": {

	    //	    System.err.println("Target: " + target.value);
	    // System.err.println("Source: " + source.value);
	       
	    if (matcher.value == null){
		matcher.value = "CosineSimilarityFingerprintMatcher";
	    }

	    if (target.value == null){
		System.err.println("Must provide target fingerprint");
		printError(CommandLineHelp.SCOREFINGERPRINTS);
	    }
		    
	    if (source.value == null){
		System.err.println("Must provide source fingerprint");
		printError(CommandLineHelp.SCOREFINGERPRINTS);
	    }


	    //todo fix fingerprint.value to accept multiple fingerprints with one argument
	    ScoreFingerprints sfp = new ScoreFingerprints(matcher.value, source.value, target.value);
	    sfp.printMatch();
	    return true;
	}
	    
	case "EXTRACTFINGERPRINTDICTIONARY": {

	    if (fingerprint.value == null) {
		printError(CommandLineHelp.EXTRACTFINGERPRINTDICTIONARY);
	    }
	    if (output.value == null) {
		ExtractFingerprintDictionary.extractDictionaryXML(fingerprint.value);
	    } else {
		ExtractFingerprintDictionary.extractDictionaryXML(fingerprint.value, output.value);
	    }
	    return true;
	}
	    
	case "COMPAREFINGERPRINT": {
	    // One to many comparison

	    if (directory.value == null){
		System.err.println("Must provide a directory to compare");
		printError(CommandLineHelp.COMPAREDIRECTORY);
	    }
		
	    if (fingerprint.value == null) {
		System.err.println("Must provide a fingerprint to compare");
		printError(CommandLineHelp.COMPAREDIRECTORY);		   
	    }

	    if(matcher.value == null ) matcher.value = "CosineSimilarityFingerprintMatcher";

	       
	    CompareFingerprint compare = new CompareFingerprint(matcher.value, fingerprint.value,
								directory.value, (int) minScore.value, null);

	    compare.compareFingerprintDirectory();
	    compare.printScores();
	    return true;
	}
	    
	case "COMPAREDIRECTORY": {
	    // N^2 comparison
	    
	    if (dictionary.value == null ) {
		System.err.println ("Missing dictionary");
		printError(CommandLineHelp.COMPAREFINGTODIR);
	    }
	    
	   
	    if (directory.value == null){
		System.err.println("Must provide a directory to compare");
		printError(CommandLineHelp.COMPAREDIRECTORY);
	    }

	    if (matcher.value == null ) matcher.value = "CosineSimilarityFingerprintMatcher";
	    
	    
	    CompareDirectory compare = new CompareDirectory(matcher.value, directory.value,
							    (int) minScore.value, dictionary.value);

	    compare.compareDirectory();
	    compare.printScores();
	    return true;
	}
		
	case "EXPERIMENT": {
	    
	    if (config.value == null){
		System.err.println("Must provide a config file");
		printError(CommandLineHelp.EXPERIMENT);
	    }
	   	    
	    Experiment exp = new Experiment(config.value);
	    File file;

	    if ((file = new File(Global.fileNameExperimentTracker)).isFile()) {
		System.out.println(Global.fileNameExperimentTracker +
				   " deleted: " + file.delete());
	    }
	    exp.runExperiment();
	    exp.aggregateResults();

	    //set the name of the graph
	    exp.graphResults(fingerprint.value);
	    return true;
	}
	    
	case "SDHASHFINGERPRINTER": {

	    if (input.value == null) printError(CommandLineHelp.SDHASHFINGERPRINTER);
	    
	    SdhashFingerprinter sdFp = new SdhashFingerprinter();
	    List<Fingerprint> list = sdFp.computeFingerprint(input.value);

	    System.out.println(list.get(0).getSdhash());
	    return true;
	}
	    
            // case "EXPERIMENTN2": {
            //     if (config.value == null) {
            //         printError(CommandLineHelp.EXPERIMENTN2);
            //     }
            //     if (matcher.value == null) {
            //         matcher.value = "CosineSimilarityFingerprintMatcher";
            //     }

            //     //todo revisit the fact that i hardcoded true
            //     ExperimentN2 exp = new ExperimentN2(directory.value, matcher.value, true);


            //     if (output != null) {
            //         exp.saveScores(output.value);
            //     } else {
            //         exp.printScores();
            //     }
            //     return true;
            // }

	default:
	    return false;
        }
    }


    /**
     * @param parser assigns field variables the parameter information passed from the CLI
     */
    void addOptions(ArgParser parser) {

        parser.addOption(getInputSpec(), input);
        parser.addOption(getOutputSpec(), output);
        parser.addOption(getDictionarySpec(), dictionary);
        parser.addOption(getConfigSpec(), config);
        parser.addOption(getDirectorySpec(), directory);
        parser.addOption(getMaxIdfSpec(), maxIDF);
        parser.addOption(getMatcherSpec(), matcher);
        parser.addOption(getMinIdfSpec(), minIDF);
        parser.addOption(getDebugSpec(), debug);
        parser.addOption(getMinScore(), minScore);
        parser.addOption(getFingerprintSpec(), fingerprint);
	parser.addOption(getSourceSpec(), source);
	parser.addOption(getTargetSpec(), target);
	

    }



    /**
     * Prints program name, use, and command line options,
     * to be used when running a program that is missing a required
     * command line option
     *
     * @param program the program to print error message for
     */
    protected static void printError(CommandLineHelp program) {

        String error = String.format("%s %n%s: %s %n%s%n", "SDText: Missing required OPTION", program.name, program.use,
				     program.options);

        ExceptionHandler.commandLineError(error);
    }


    /**
     * Prints SDText help screen-- this is all of the programs at once/and has their options
     */
    protected static void printHelp() {
        boolean print_programs = false;
        for (CommandLineHelp program : CommandLineHelp.values()) {
            if (print_programs) {
                System.err.println();
                System.err.println("PROGRAMS:");
                System.err.println();
                print_programs = false;
            }
            System.err.println(program.name + ": " + program.use);
            System.err.println(program.options);
            System.err.println();
            if (program == CommandLineHelp.HELP)
                print_programs = true;
        }
    }


}
